@@ -434,7 +434,7 @@ on_setup_io(void *data, int fd, uint32_t mask)
 		if (len < (int)sizeof(params))
 			goto short_packet;
 
-		if (strcmp(params.type, "params") != 0)
+		if (strncmp(params.type, "params", sizeof(params.type)) != 0)
 			goto wrong_type;
 
 		switch(ntohl(params.packet_id)) {
@@ -1072,7 +1072,7 @@ on_socket_io(void *data, int fd, uint32_t mask)
 		if (len < (int)sizeof(params))
 			goto short_packet;
 
-		if (strcmp(params.type, "params") != 0)
+		if (strncmp(params.type, "params", sizeof(params.type)) != 0)
 			goto wrong_type;
 
 		switch(ntohl(params.packet_id)) {
